from typing import Dict, Union, List, Tuple
import tensorflow as tf
from connectionist.layers import PMSPLayer, HNSLayer
from connectionist.surgery import SurgeryPlan, Surgeon, make_recipient, copy_transplant


class PMSP(tf.keras.Model):
    """PMSP sim 3 model. A recurrent neural network with hidden layer, phonology layer and cleanup layer. It takes orthography as input and outputs phonology.

    See Plaut, McClelland, Seidenberg and Patterson (1996), simulation 3 for details.

    Args:
        tau: the time constant (smaller time constant = slower temporal dynamics, tau in [0, 1])
        h_units: the number of hidden units
        p_units: the number of phonology units
        c_units: the number of cleanup units
        connections: the connections between layers, default to: ['oh', 'ph', 'hp', 'pp', 'cp', 'pc']

    Call args:

        inputs: the input tensor, shape: (batch_size, time_steps, o_units)
        return_internals: whether to return the internal dynamics, default to: False.

            If `return_internals`=`True`, the model will return a dictionary with all the internal dynamics. Including:
                - activations in all layers (e.g., 'hidden')
                - raw inputs in each connection (e.g., 'oh' = $o @ w_oh$)

            If `return_internals`=`False` the model will return the phonology activation, i.e., $p$.

    Example:

    ```python
    import tensorflow as tf
    from connectionist.data import ToyOP
    from connectionist.models import PMSP

    data = ToyOP()
    model = PMSP(tau=0.2, h_units=10, p_units=9, c_units=5)
    model.compile(
        optimizer=tf.keras.optimizers.Adam(),
        loss=tf.keras.losses.BinaryCrossentropy(),
    )
    model.fit(data.x_train, data.y_train, epochs=3, batch_size=20)
    ```

    """

    def __init__(
        self,
        tau: float,
        h_units: int,
        p_units: int,
        c_units: int,
        h_noise: float = 0.0,
        p_noise: float = 0.0,
        c_noise: float = 0.0,
        connections: List[str] = None,
        l2: float = 0.0,
    ) -> None:
        super().__init__()

        self.tau = tau
        self.h_units = h_units
        self.p_units = p_units
        self.c_units = c_units
        self.h_noise = h_noise
        self.p_noise = p_noise
        self.c_noise = c_noise
        self.connections = connections
        self.l2 = l2

        if connections is None:
            self.connections = ["oh", "ph", "hp", "pp", "cp", "pc"]

        self.pmsp = PMSPLayer(
            tau=tau,
            h_units=self.h_units,
            p_units=self.p_units,
            c_units=self.c_units,
            h_noise=self.h_noise,
            p_noise=self.p_noise,
            c_noise=self.c_noise,
            connections=self.connections,
            l2=self.l2,
        )

    @property
    def all_layers_names(self) -> List[str]:
        return self.pmsp.all_layers_names

    @property
    def connection_locs(self) -> Dict[str, Dict[str, int]]:
        """A map that shows which axis of a weight is connecting to a layer."""

        return {layer: self._find_conn_locs(layer) for layer in self.all_layers_names}

    @property
    def weights_abbreviations(self) -> Dict[str, str]:
        """Weights name abbreviations. (e.g., w_oh, bias_hidden)."""

        weights = {f"w_{w}": f"{w}/kernel:0" for w in self.connections}

        biases = {f"bias_{layer}": f"{layer}/bias:0" for layer in self.all_layers_names}
        return {**weights, **biases}

    def call(
        self, inputs: tf.Tensor, training: bool = False, return_internals: bool = False
    ) -> Union[tf.Tensor, Dict[str, tf.Tensor]]:
        return self.pmsp(inputs, training=training, return_internals=return_internals)

    def get_config(self) -> Dict[str, Union[float, int, list]]:
        return dict(
            tau=self.tau,
            h_units=self.h_units,
            p_units=self.p_units,
            c_units=self.c_units,
            h_noise=self.h_noise,
            p_noise=self.p_noise,
            c_noise=self.c_noise,
            connections=self.connections,
        )

    # Miscellaneous internal methods

    def _find_conn_locs(self, layer: str) -> Dict[str, int]:
        """Find the connection locations of the target layer."""

        def find(l: List[str], prefix: str):
            """Locate connection by layer prefix."""
            return [i for i, x in enumerate(l) if x == prefix]

        prefix = layer[0]
        conn_weights = {}
        for connection in self.connections:
            if prefix in connection:
                conn_weights[f"w_{connection}"] = find(connection, prefix)

        conn_bias = {f"bias_{layer}": [0]}
        return {**conn_weights, **conn_bias}

    def _validate_layer(self, layer: str) -> None:
        """Validate the target layer."""

        if layer not in self.all_layers_names:
            raise ValueError(
                f"Unknown target layer: {layer}, please choose from {self.all_layers_names}"
            )

    def _validate_connections(self, connections: List[str]) -> None:
        if not all([c in self.connections for c in connections]):
            raise ValueError(
                f"Unknown connections: {connections}, please choose from {self.connections}"
            )

    def _to_units(self, layer: str) -> int:
        """Get the number of units in the target layer.

        Args:
            layer: the target layer, choose from ['hidden', 'phonology', 'cleanup']
        """

        mapping = {
            "hidden": self.h_units,
            "phonology": self.p_units,
            "cleanup": self.c_units,
        }
        return mapping[layer]

    # Extra methods for damaging the model

    def shrink_layer(self, layer: str, rate: float) -> tf.keras.Model:
        """Shrink the weights of the target layer.

        Args:
            layer: the target layer, choose from ['hidden', 'phonology', 'cleanup']
            rate: the shrink rate

        Returns:
            A new model with the same architecture, but with new weights shapes that match with the shrank layer.

        Example:

        ```python
        from connectionist.data import ToyOP
        from connectionist.models import PMSP

        model = PMSP(tau=0.2, h_units=10, p_units=9, c_units=5)

        # Build or call or fit a model to instantiate the weights
        y = model(data.x_train)

        new_model = model.shrink_layer('hidden', rate=0.5)
        ```

        """

        self._validate_layer(layer)

        plan = SurgeryPlan(
            layer=layer, original_units=self._to_units(layer), shrink_rate=rate
        )
        surgeon = Surgeon(surgery_plan=plan)

        # Make a new model with the same architecture, but with new weights shapes
        new_model = make_recipient(model=self, surgery_plan=plan, make_model_fn=PMSP)
        new_model.build(input_shape=self.pmsp._build_input_shape)

        surgeon.transplant(donor=self, recipient=new_model)
        return new_model

    def _build_and_transplant(self, config: dict) -> tf.keras.Model:
        """Build a new model with the same architecture with new config."""

        new_model = PMSP(**config)
        new_model.build(input_shape=self.pmsp._build_input_shape)

        # Copy weights
        for weight_name in new_model.weights_abbreviations.keys():
            copy_transplant(donor=self, recipient=new_model, weight_name=weight_name)

        return new_model

    def cut_connections(self, connections: List[str]) -> tf.keras.Model:
        """Cut connections between two layers.

        Args:
            connections: the connections to be cut, the connections most be found in the original model, i.e., in `model.connections`.

        Returns:
            A new model with the same architecture, but with new connections.

        """

        self._validate_connections(connections)

        # Create recipient model with less connections
        model_config = self.get_config()
        remaining_connections = [c for c in self.connections if c not in connections]
        model_config.update(connections=remaining_connections)
        return self._build_and_transplant(model_config)

    def add_noise(self, layer: str, stddev: float) -> tf.keras.Model:
        """Add noise to the target layer.

        The noise is active in both training and inference.

        Args:
            layer: the target layer, choose from ['hidden', 'phonology', 'cleanup']
            stddev: the standard deviation of the noise

        Example:

        ```python
        from connectionist.data import ToyOP
        from connectionist.models import PMSP

        model = PMSP(tau=0.2, h_units=10, p_units=9, c_units=5)

        # Build or call or fit a model to instantiate the weights
        y = model(data.x_train)

        new_model = model.add_noise('hidden', stddev=0.1)
        ```

        """

        self._validate_layer(layer)

        # Create recipient model with noise at the target layer
        model_config = self.get_config()

        def _to_noise_name(layer_name: str) -> str:
            return f"{layer_name[0]}_noise"

        noise_name = _to_noise_name(layer)
        model_config[noise_name] = model_config[noise_name] + stddev
        return self._build_and_transplant(model_config)

    def add_l2(self, l2: float) -> tf.keras.Model:
        """Add L2 regularization to all the weights and biases in the model."""

        model_config = self.get_config()
        model_config.update(l2=l2)
        return self._build_and_transplant(model_config)


class HubAndSpokes(tf.keras.Model):
    def __init__(
        self,
        tau: float,
        hub_name: str,
        hub_units: int,
        spoke_names: List[str],
        spoke_units: List[int],
    ) -> None:

        super().__init__()

        self.tau = tau
        self.hub_name, self.hub_units = hub_name, hub_units
        self.spoke_names, self.spoke_units = spoke_names, spoke_units

        self.hns = HNSLayer(
            tau=tau,
            hub_name=hub_name,
            hub_units=hub_units,
            spoke_names=spoke_names,
            spoke_units=spoke_units,
        )

    def call(self, x: Dict[str, tf.Tensor]) -> Dict[str, tf.Tensor]:
        return self.hns(x)

    def train_step(self, data: Tuple[Dict[str, tf.Tensor]]) -> Dict[str, tf.Tensor]:
        """Train the model for one step, return losses and metrics."""

        x, y = data

        with tf.GradientTape() as tape:
            y_pred = self(x, training=True)

            all_losses = []

            # Only inject error according to y_train keys
            for y_name, y_target in y.items():
                sum_loss = tf.reduce_sum(
                    self.compiled_loss(y_true=y_target, y_pred=y_pred[y_name])
                )  # reduce over batch_size axis
                all_losses.append(sum_loss)

            loss_value = tf.reduce_sum(
                all_losses
            )  # Final loss value is the grand sum over every axis (output layer, batch size, time ticks, units).

        grads = tape.gradient(
            loss_value, self.trainable_weights
        )  # compute gradients dL/dw
        self.optimizer.apply_gradients(
            zip(grads, self.trainable_weights)
        )  # update weights using stock optimizer

        self.compiled_metrics.update_state(y, y_pred)  # update metrics
        return {m.name: m.result() for m in self.metrics}



from typing import Dict, List, Optional, Tuple, Union
from functools import partial
import tensorflow as tf


def _time_averaging(
    x: tf.Tensor, tau: float, states: Optional[tf.Tensor] = None
) -> tf.Tensor:
    """Time-averaging mechanism.

    Args:
        x (tf.Tensor): Input tensor.
        tau (float): Time-averaging parameter (How much information should take from the new input). range: [0, 1].
        states (tf.Tensor, optional): Last states (last activation: a_{t-1} or last integrated input: x_{t-1} = \sum ). Defaults to None.
    """

    if states is None:
        return x * tau

    return x * tau + (1 - tau) * states


def reshape_proper(a: tf.TensorArray, perm: List[int] = None) -> tf.TensorArray:
    """Reshape the TensorArray to the standard output shape of [batch_size, sequence_length, feature].

    Args:
        a (tf.TensorArray): TensorArray to be reshaped.
        perm: Permutation of the dimensions of the input TensorArray. Defaults to [1, 0, 2] for typical RNN use case.
    """
    if perm is None:
        perm = [1, 0, 2]
    return tf.transpose(a.stack(), perm)


class TimeAveragedDense(tf.keras.layers.Dense):
    """Dense layer with Time-averaging mechanism.

    In short, time-averaging mechanism simulates continuous-temporal dynamics in a discrete-time recurrent neural networks.
    See Plaut, McClelland, Seidenberg, and Patterson (1996) equation (15) for more details.

    Args:
        tau (float): Time-averaging parameter (How much information should take from the new input). range: [0, 1].

        average_at (str): Where to average. Options: 'before_activation', 'after_activation'.

            When average_at is 'before_activation', the time-averaging is applied BEFORE activation. i.e., time-averaging INPUT.:
                outputs = activation(integrated_input);
                integrated input = tau * (inputs @ weights + bias) + (1-tau) * last_integrated_input;
                last_integrated_input is obtained from the last call of this layer, its values stored at `self.states`

            When average_at is 'after_activation', the time-averaging is applied AFTER activation. i.e., time-averaging OUTPUT.:
                outputs = tau * activation(inputs @ weights + bias) + (1-tau) * last_outputs;
                last_outputs is obtained from the last call of this layer, its values stored at `self.states`

        kwargs (str, optional): Any argument in keras.layers.Dense. (e.g., units, activation, use_bias, kernel_initializer, bias_initializer,
            kernel_regularizer, bias_regularizer, activity_regularizer, kernel_constraint, bias_constraint, **kwargs).

    """

    def __init__(
        self,
        tau: float,
        average_at: str,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.tau = tau
        self.average_at = average_at

        # Turn off activation, will manually call the ._activation in `.call()`.
        self._activation = self.activation
        self.activation = None

        if self.tau < 0 or self.tau > 1:
            raise ValueError(f"tau must be between 0 and 1, but got {self.tau}")

        if self.average_at not in ["before_activation", "after_activation"]:
            raise ValueError(
                f"average_at must be one of ['before_activation', 'after_activation'], but got {self.average_at}"
            )

        self.states = None

    def call(self, inputs: tf.Tensor) -> tf.Tensor:

        # keras.layers.Dense.call() without activation
        outputs = super().call(inputs)
        if self.average_at == "before_activation":
            outputs = _time_averaging(outputs, self.tau, self.states)
            self.states = outputs  # state is integrated input here

        if self._activation is not None:
            outputs = self._activation(outputs)

        if self.average_at == "after_activation":
            outputs = _time_averaging(outputs, self.tau, self.states)
            self.states = outputs  # state is activation here

        return outputs

    def reset_states(self) -> None:
        self.states = None

    def get_config(self) -> dict:
        config = super().get_config()
        config.update(
            {
                "tau": self.tau,
                "average_at": self.average_at,
                "activation": tf.keras.activations.serialize(self._activation),
            }
        )
        return config


class MultiInputTimeAveraging(tf.keras.layers.Layer):
    """Time-averaging mechanism for multiple inputs.

    In short, time-averaging mechanism simulates continuous-temporal dynamics in a discrete-time recurrent neural networks.
    See Plaut, McClelland, Seidenberg, and Patterson (1996) equation (15) for more details.

    This layer is designed for multiple inputs, assuming they had ALREADY been multiplied by weights i.e., a list of (x @ w).
    i.e., there is only bias term in this layer (no weights) if use_bias is True.

    Args:
        tau (float): Time-averaging parameter (How much information should take from the new input). range: [0, 1].

        average_at (str): Where to average. Options: 'before_activation', 'after_activation'.

            When average_at is 'before_activation', the time-averaging is applied BEFORE activation. i.e., time-averaging INPUT.:
                outputs = activation(integrated_input);
                integrated input = tau * (sum(inputs) + bias) + (1-tau) * last_integrated_input;
                last_integrated_input is obtained from the last call of this layer, its values stored at `self.states`

            When average_at is 'after_activation', the time-averaging is applied AFTER activation. i.e., time-averaging OUTPUT.:
                outputs = tau * activation(sum(inputs) + bias) + (1-tau) * last_outputs;
                last_outputs is obtained from the last call of this layer, its values stored at `self.states`

        activation (str, optional): Activation function to use. Defaults to None.

        use_bias (bool, optional): Whether the layer uses a bias vector. Defaults to True.

        bias_initializer (optional): Initializer for the bias vector. Defaults to 'zeros'.

        bias_regularizer (optional): Regularizer function applied to the bias vector. Defaults to None.

        bias_constraint (optional): Constraint function applied to the bias vector. Defaults to None.

    """

    def __init__(
        self,
        tau: float,
        average_at: str,
        activation: str = None,
        use_bias: bool = True,
        bias_initializer: str = "zeros",
        bias_regularizer: str = None,
        bias_constraint: str = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.tau = tau
        self.average_at = average_at
        self.activation = tf.keras.activations.get(activation)
        self.use_bias = use_bias
        self.bias_initializer = tf.keras.initializers.get(bias_initializer)
        self.bias_regularizer = tf.keras.regularizers.get(bias_regularizer)
        self.bias_constraint = tf.keras.constraints.get(bias_constraint)

        if self.tau < 0 or self.tau > 1:
            raise ValueError(f"tau must be between 0 and 1, but got {self.tau}")

        if self.average_at not in ["before_activation", "after_activation"]:
            raise ValueError(
                f"average_at must be one of ['before_activation', 'after_activation'], but got {self.average_at}"
            )

        self.states = None

    def build(self, input_shape) -> None:

        self.sum = tf.keras.layers.Add()

        if self.use_bias:
            self.bias = self.add_weight(
                "bias",
                shape=[
                    input_shape[0][-1],
                ],
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                trainable=True,
            )
        else:
            self.bias = None

        self.built = True

    def call(self, inputs: List[tf.Tensor]) -> tf.Tensor:

        # Check if all inputs have the same shape
        if not all([x.shape == inputs[0].shape for x in inputs]):
            raise ValueError("All inputs must have the same shape.")

        if len(inputs) == 1:
            outputs = inputs[0]
        else:
            outputs = self.sum(inputs)

        if self.use_bias:
            outputs = tf.nn.bias_add(outputs, self.bias)

        if self.average_at == "before_activation":
            outputs = _time_averaging(outputs, self.tau, self.states)
            self.states = outputs  # state is integrated input here

        if self.activation is not None:
            outputs = self.activation(outputs)

        if self.average_at == "after_activation":
            outputs = _time_averaging(outputs, self.tau, self.states)
            self.states = outputs  # state is activation here

        return outputs

    def reset_states(self) -> None:
        self.states = None

    def get_config(self) -> dict:
        config = super().get_config()
        config.update(
            {
                "tau": self.tau,
                "average_at": self.average_at,
                "units": self.units,
                "activation": tf.keras.activations.serialize(self.activation),
                "use_bias": self.use_bias,
                "bias_initializer": tf.keras.initializers.serialize(
                    self.bias_initializer
                ),
                "bias_regularizer": tf.keras.regularizers.serialize(
                    self.bias_regularizer
                ),
                "bias_constraint": tf.keras.constraints.serialize(self.bias_constraint),
            }
        )
        return config


class TimeAveragedRNNCell(tf.keras.layers.Layer):
    def __init__(self, tau: float, units: int) -> None:
        super().__init__()
        self.tau = tau
        self.units = units

    def build(self, input_shape: tf.TensorShape) -> None:
        self.input_dense = tf.keras.layers.Dense(self.units, use_bias=False)  # w_xh
        self.recurrent_dense = tf.keras.layers.Dense(self.units, use_bias=False)  # w_hh
        self.time_averaging = MultiInputTimeAveraging(
            tau=self.tau, average_at="after_activation", activation="sigmoid"
        )  # "time-averaging mechanism" with multiple inputs
        self.built = True

    def call(self, inputs: tf.Tensor, states=None) -> Tuple[tf.Tensor, tf.Tensor]:
        xh = self.input_dense(inputs)  # x @ w_xh

        if states is None:
            hh = tf.zeros_like(xh)
        else:
            hh = self.recurrent_dense(states)  # h @ w_hh

        outputs = self.time_averaging(
            [xh, hh]
        )  # sigmoid (tau * (xh + hh + bias) + (1 - tau) * last activation)
        return (
            outputs,
            outputs,
        )  # Consistent with the RNN API, one for state and one for output

    def reset_states(self) -> None:
        self.time_averaging.reset_states()  # Reset the states of the time-averaging mechanism (last activation = None)


class TimeAveragedRNN(tf.keras.layers.Layer):
    def __init__(self, tau: float, units: int) -> None:
        super().__init__()
        self.tau = tau
        self.units = units

    def build(self, input_shape: tf.TensorShape) -> None:
        self.rnn_cell = TimeAveragedRNNCell(tau=self.tau, units=self.units)
        self.built = True

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        max_ticks = inputs.shape[1]  # (batch_size, seq_len, input_dim)
        outputs = tf.TensorArray(dtype=tf.float32, size=max_ticks)
        states = None

        for t in range(max_ticks):
            this_tick_input = inputs[:, t, :]
            states, output = self.rnn_cell(this_tick_input, states=states)
            outputs = outputs.write(t, output)

        # rnn_cell states persist across ticks, but not across batches/calls, so we need to reset rnn_cell here
        self.rnn_cell.reset_states()
        return reshape_proper(outputs)


class PMSPCell(tf.keras.layers.Layer):
    """RNN cell for PMSP model.

    See Plaut, McClelland, Seidenberg and Patterson (1996), simulation 3.
    """

    def __init__(
        self,
        tau: float,
        h_units: int,
        p_units: int,
        c_units: int,
        h_noise: float = 0.0,
        p_noise: float = 0.0,
        c_noise: float = 0.0,
        connections: List[str] = None,
        l2: float = 0.0,
    ) -> None:

        super().__init__()
        self.tau = tau
        self.h_units = h_units
        self.p_units = p_units
        self.c_units = c_units

        [self._validate_noise(x) for x in [h_noise, p_noise, c_noise]]
        self.h_noise = h_noise
        self.p_noise = p_noise
        self.c_noise = c_noise

        if connections is not None:
            self._validate_connections(connections)
        else:
            connections = ["oh", "ph", "hp", "pp", "cp", "pc"]

        self.connections = connections

        self.l2 = l2

    @property
    def all_layers_names(self) -> List[str]:
        return [
            "hidden",
            "phonology",
            "cleanup",
        ]  # Prefixes ('h', 'p', 'c') must be unique

    @staticmethod
    def _validate_connections(connections) -> None:
        s = set([letter for connection in connections for letter in connection])

        if not s.issubset(set("ohpc")):
            raise ValueError(
                "Connections must contain only letters in ['o', 'h', 'p', 'c']"
            )

    @staticmethod
    def _validate_noise(noise: float) -> None:
        if noise < 0.0:
            raise ValueError("Noise must be > 0")

    def get_connection_units(self, connection: str) -> int:

        self._validate_connections([connection])
        target = connection[-1]
        return {
            "h": self.h_units,
            "p": self.p_units,
            "c": self.c_units,
        }[target]

    def build(self, input_shape: tf.TensorShape) -> None:

        regularizer = tf.keras.regularizers.L2(self.l2)

        for connection in self.connections:
            setattr(
                self,
                connection,
                tf.keras.layers.Dense(
                    units=self.get_connection_units(connection),
                    use_bias=False,
                    name=connection,
                    kernel_regularizer=regularizer,
                    bias_regularizer=regularizer,
                ),
            )

        # layer block: Add bias, time-averaging and activation
        create_layer = partial(
            MultiInputTimeAveraging,
            tau=self.tau,
            average_at="after_activation",
            activation="sigmoid",
            bias_regularizer=regularizer,  # Only have bias in MITA layer
        )

        for layer in self.all_layers_names:
            setattr(self, layer, create_layer(name=layer))

        # Noise layers (in dict for easier access)
        def _get_noise(layer_name: str) -> float:
            return getattr(self, f"{layer_name[0]}_noise")

        self.noise = {}
        for layer_name in self.all_layers_names:
            self.noise[layer_name] = tf.keras.layers.GaussianNoise(
                _get_noise(layer_name)
            )

        self.built = True

    def get_connections(self, layer: str) -> List[str]:
        """Get connections that end with the given layer.

        e.g., if layer = "hidden", return all connections that ends with "h", e.g.: ["oh", "ph"]
        """
        return [conn for conn in self.connections if conn.endswith(layer[0])]

    def _has_connection(self, layer: str) -> bool:
        """Check whether the given layer has any incoming connections."""
        return len(self.get_connections(layer)) > 0

    def call(
        self,
        last_o: tf.Tensor,
        last_h: tf.Tensor,
        last_p: tf.Tensor,
        last_c: tf.Tensor,
        training: bool = False,
        return_internals: bool = False,
    ) -> Dict[str, tf.Tensor]:
        def get_input(connection) -> tf.Tensor:
            input_map = {
                "o": last_o,
                "h": last_h,
                "p": last_p,
                "c": last_c,
            }
            return input_map[connection[0]]

        batch_size = last_o.shape[0]
        layer_activations = {}
        inputs_to = {}
        for layer_name in self.all_layers_names:
            inputs_to[layer_name] = {}  # Layer inputs, e.g.: noise, x @ w_{xy}

            # Add noise (I did not separate 0 noise from non-zero noise for simplicity, and it is computationally cheap anyway)
            # Do NOT rely on shape broadcasting, since each value must have a random noise value
            _zeros = tf.zeros((batch_size, getattr(self, f"{layer_name[0]}_units")))

            # I use `f"{layer_name}_noise"` instead of `noise` because `input_to` will be flattened at the end
            # Training is always True, because we want to add noise even during testing
            inputs_to[layer_name][f"{layer_name}_noise"] = self.noise[layer_name](
                _zeros, training=True
            )

            # Append x @ w for each incoming connection
            if self._has_connection(layer_name):
                for conn in self.get_connections(layer_name):
                    inputs_to[layer_name][conn] = getattr(self, conn)(get_input(conn))

            # Layer activation (bias is inside the MultiInputTimeAveraging layer, so we don't need to add it here)
            layer_activations[layer_name] = getattr(self, layer_name)(
                inputs_to[layer_name].values()
            )

        if return_internals:
            return {
                **layer_activations,
                **inputs_to["hidden"],
                **inputs_to["phonology"],
                **inputs_to["cleanup"],
            }

        return layer_activations

    def reset_states(self):
        """Reset time averaging history."""

        for layer in self.all_layers_names:
            getattr(self, layer).reset_states()


class PMSPLayer(tf.keras.layers.Layer):
    """PMSP sim 3 model's layer.

    See Plaut, McClelland, Seidenberg and Patterson (1996), simulation 3.
    """

    def __init__(
        self,
        tau: float,
        h_units: int,
        p_units: int,
        c_units: int,
        h_noise: float = 0.0,
        p_noise: float = 0.0,
        c_noise: float = 0.0,
        connections: List[str] = None,
        l2: float = 0.0,
    ) -> None:
        super().__init__()

        # Sanitation is done when building the cell to avoid duplication
        self.tau = tau
        self.h_units, self.p_units, self.c_units = h_units, p_units, c_units
        self.h_noise, self.p_noise, self.c_noise = h_noise, p_noise, c_noise
        self.connections = connections
        self.l2 = l2

    def build(self, input_shape: tf.TensorShape) -> None:
        self.cell = PMSPCell(
            tau=self.tau,
            h_units=self.h_units,
            p_units=self.p_units,
            c_units=self.c_units,
            h_noise=self.h_noise,
            p_noise=self.p_noise,
            c_noise=self.c_noise,
            connections=self.connections,
            l2=self.l2,
        )

        # Lifting `all_layers_name` from cell to layer for easier access
        self.connections = self.cell.connections  # Lift instantiated connections
        self.all_layers_names = self.cell.all_layers_names

        self.built = True

    def call(
        self,
        inputs: tf.Tensor,
        training: bool = False,
        return_internals: bool = False,
    ) -> Union[tf.Tensor, Dict[str, tf.Tensor]]:

        batch_size, max_ticks, _ = inputs.shape

        output_names = (
            [*self.all_layers_names, *self.connections]
            if return_internals
            else ["phonology"]
        )

        # Containers for outputs with shape (batch_size, max_ticks, units)
        tf_arrays = {
            name: tf.TensorArray(dtype=tf.float32, size=max_ticks)
            for name in output_names
        }

        # Initialize layer activations
        h = tf.zeros((batch_size, self.h_units))
        p = tf.zeros((batch_size, self.p_units))
        c = tf.zeros((batch_size, self.c_units))

        # Run RNN (Unrolling RNN Cell)
        for t in range(max_ticks):
            o = tf.cast(inputs[:, t], tf.float32)

            cell_outputs = self.cell(
                last_o=o,
                last_h=h,
                last_p=p,
                last_c=c,
                training=training,
                return_internals=return_internals,
            )
            h, p, c = (
                cell_outputs["hidden"],
                cell_outputs["phonology"],
                cell_outputs["cleanup"],
            )

            # Store output arrays
            for name in output_names:
                tf_arrays[name] = tf_arrays[name].write(t, cell_outputs[name])

        self.cell.reset_states()

        return {name: reshape_proper(tf_arrays[name]) for name in output_names}


class Spoke(tf.keras.layers.Layer):
    """A spoke in the hub-and-spokes model."""

    def __init__(
        self,
        tau: float,
        units: int,
        average_at="after_activation",
        activation="sigmoid",
        **kwargs,
    ) -> None:
        super().__init__()
        self.tau = tau
        self.units = units  # Begin explicit to cater for `None` input (which is possible in the hub-and-spokes model)
        self.time_averaging = MultiInputTimeAveraging(
            tau=self.tau, average_at=average_at, activation=activation, **kwargs
        )

    def call(
        self, inputs: tf.Tensor = None, cross_tick_states: List[tf.Tensor] = None
    ) -> tf.Tensor:  # to avoid confusing with `self.time_averaging.states` (a_{t-1}), I use a new name `cross_tick_states` here to represent the cross ticks connection.
        """Call the spoke.

        Args:
            inputs: clamped input
            cross_tick_states: states from the red arrows (cross ticks projection), a_i w_{ij}.
        """
        if isinstance(cross_tick_states, list):
            if len(cross_tick_states) == 0:
                cross_tick_states = None

        if inputs is None and cross_tick_states is None:
            return self.time_averaging([tf.zeros((1, self.units))])

        net_inputs = []

        if inputs is not None:
            net_inputs.append(inputs)

        if cross_tick_states is not None:
            net_inputs.extend(cross_tick_states)

        return self.time_averaging(net_inputs)

    def reset_states(self):
        self.time_averaging.reset_states()


class HNSCell(tf.keras.layers.Layer):
    def __init__(
        self,
        tau: float,
        hub_name: str,
        hub_units: int,
        spoke_names: List[str],
        spoke_units: List[int],
    ) -> None:
        super().__init__()
        self.tau = tau
        self.hub_name = hub_name
        self.hub_units = hub_units
        self.spoke_names = spoke_names
        self.spoke_units = spoke_units

    def build(self, input_shape) -> None:

        # Hub
        self.hub = MultiInputTimeAveraging(
            tau=self.tau,
            average_at="after_activation",
            activation="sigmoid",
            use_bias=True,
            name=self.hub_name,
        )
        self.w_hh = self.add_weight(
            shape=(self.hub_units, self.hub_units),
            initializer="random_normal",
            trainable=True,
            name="w_hh",
        )

        # Spokes
        for i, (name, units) in enumerate(zip(self.spoke_names, self.spoke_units)):
            setattr(self, name, Spoke(tau=self.tau, units=units, name=name))

            # Outgoing connection weights (w_sih)
            setattr(
                self,
                f"w_s{i}h",
                self.add_weight(
                    shape=(units, self.hub_units),
                    initializer="random_normal",
                    trainable=True,
                    name=f"w_s{i}h",
                ),
            )

            # Self connection weights (w_sisi)
            setattr(
                self,
                f"w_s{i}s{i}",
                self.add_weight(
                    shape=(units, units),
                    initializer="random_normal",
                    trainable=True,
                    name=f"w_s{i}s{i}",
                ),
            )

            # Incoming connection weights (w_hsi)
            setattr(
                self,
                f"w_hs{i}",
                self.add_weight(
                    shape=(self.hub_units, units),
                    initializer="random_normal",
                    trainable=True,
                    name=f"w_hs{i}",
                ),
            )

        self.built = True

    def _validate_spoke_x(
        self, x: Optional[Dict[str, tf.Tensor]]
    ) -> Dict[str, Optional[tf.Tensor]]:
        """Validate and clean inputs and last_act_spokes."""

        if x is None:
            return {name: None for name in self.spoke_names}

        for name in self.spoke_names:
            if name not in x.keys():
                x[name] = None

        for name in x.keys():
            if name not in self.spoke_names:
                raise ValueError(
                    f"{name} is not one of the spoke name: {self.spoke_names}."
                )

        return x

    def call(
        self,
        inputs: Optional[Dict[str, tf.Tensor]] = None,
        last_act_hub: tf.Tensor = None,
        last_act_spokes: Optional[Dict[str, tf.Tensor]] = None,
    ) -> Tuple[tf.Tensor, Dict[str, tf.Tensor]]:
        """Call the HNSCell.

        Args:
            inputs (dict): clamped inputs with spoke names as keys.
            last_act_hub (tf.Tensor): last activation of the hub.
            last_act_spokes (dict): last activations of the spokes with spoke names as keys.

        Returns:
            act_hub (tf.Tensor): activation of the hub.
            act_spokes (dict): activations of the spokes with spoke names as keys.
        """

        inputs = self._validate_spoke_x(inputs)

        last_act_spokes = self._validate_spoke_x(last_act_spokes)

        act_spokes = {}
        inputs_to_hub = []

        # Calculate spokes activations
        for i, name in enumerate(self.spoke_names):
            cross_tick_states = []

            if last_act_hub is not None:  # From Hub
                cross_tick_states.append(last_act_hub @ getattr(self, f"w_hs{i}"))

            if last_act_spokes[name] is not None:  # From self-connection
                cross_tick_states.append(
                    last_act_spokes[name] @ getattr(self, f"w_s{i}s{i}")
                )

            act_spokes[name] = getattr(self, name)(inputs[name], cross_tick_states)

            # Also append the inputs to hub
            inputs_to_hub.append(act_spokes[name] @ getattr(self, f"w_s{i}h"))

        # calculate hub activation
        if last_act_hub is not None:
            inputs_to_hub.append(last_act_hub @ self.w_hh)

        act_hub = self.hub(inputs_to_hub)
        return act_hub, act_spokes

    def reset_states(self) -> None:
        self.hub.reset_states()
        for name in self.spoke_names:
            getattr(self, name).reset_states()


class HNSLayer(tf.keras.layers.Layer):
    def __init__(
        self,
        tau: float,
        hub_name: str,
        hub_units: int,
        spoke_names: List[str],
        spoke_units: List[int],
    ) -> None:
        super().__init__()
        self.tau = tau
        self.hub_name, self.hub_units = hub_name, hub_units
        self.spoke_names, self.spoke_units = spoke_names, spoke_units

    def build(self, input_shape) -> None:
        self.cell = HNSCell(
            tau=self.tau,
            hub_name=self.hub_name,
            hub_units=self.hub_units,
            spoke_names=self.spoke_names,
            spoke_units=self.spoke_units,
        )
        self.built = True

    @staticmethod
    def _get_batch_size_and_max_tick(inputs: Dict[str, tf.Tensor]) -> Tuple[int]:
        """Get the batch size and max ticks from inputs."""
        for x in inputs.values():
            if x is not None:
                return x.shape[0], x.shape[1]
        raise ValueError("No input is given, cannot infer batch size or max ticks.")

    def call(self, inputs: Dict[str, tf.Tensor]) -> List[tf.Tensor]:

        inputs = self.cell._validate_spoke_x(inputs)
        batch_size, max_ticks = self._get_batch_size_and_max_tick(inputs)

        # Initialize activations in hub and spokes
        hub = tf.zeros((batch_size, self.hub_units))
        spokes = {}
        for name, units in zip(self.spoke_names, self.spoke_units):
            spokes[name] = tf.zeros((batch_size, units))

        # Make containers for outputs
        all_names = [self.hub_name, *self.spoke_names]
        tf_arrays = {
            name: tf.TensorArray(tf.float32, size=max_ticks) for name in all_names
        }

        # Unrolling
        for t in range(max_ticks):
            hub, spokes = self.cell(
                inputs={name: v[:, t] for name, v in inputs.items() if v is not None},
                last_act_hub=hub,
                last_act_spokes=spokes,
            )

            # Write hub and spokes activation to the container
            tf_arrays[self.hub_name] = tf_arrays[self.hub_name].write(t, hub)
            for name, spoke in spokes.items():
                tf_arrays[name] = tf_arrays[name].write(t, spoke)

        self.cell.reset_states()

        return {name: reshape_proper(tf_arrays[name]) for name in all_names}



import tensorflow as tf


class MaskedBinaryCrossEntropy(tf.keras.losses.Loss):
    """Compute Binary Cross-Entropy with mask.

    Args:
        y_true: target y with shape (batch_size, seq_len, feature)
        y_pred: predicted y with shape (batch_size, seq_len, feature)
        mask_value: value in y_true to be masked
    Returns:
        Loss value with shape (batch_size)
    """

    def __init__(
        self,
        mask_value: int = None,
        name="masked_binary_crossentropy",
        reduction="none",
        **kwargs
    ) -> None:
        super().__init__(name=name, reduction=reduction, **kwargs)
        self.mask_value = mask_value

    def call(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        epsilon = tf.keras.backend.epsilon()  # very small value to avoid log(0)
        cross_entropy = y_true * tf.math.log(y_pred + epsilon)
        cross_entropy = cross_entropy + (1 - y_true) * tf.math.log(1 - y_pred + epsilon)

        if self.mask_value:
            mask = tf.cast(
                tf.where(y_true == self.mask_value, 0, 1), tf.float32
            )  # create mask
        else:
            mask = tf.ones_like(cross_entropy)  # All inclusive mask if value is none

        cross_entropy = mask * cross_entropy  # zero out the masked values
        cross_entropy = tf.reduce_sum(
            cross_entropy, axis=[1, 2]
        )  # sum over all units (axis 2) and time steps (axis 1)
        return -cross_entropy / (
            epsilon + tf.reduce_sum(mask, axis=[1, 2])
        )  # - (1/N) sum(y * log(p(y)) + (1-y) * log(1-p(y)))



import random
from dataclasses import dataclass
from typing import Callable, List
import tensorflow as tf


def check_shapes(donor: tf.keras.Model, recipient: tf.keras.Model) -> None:
    """Check that the shapes of the weights are the same."""
    for w_donor, w_recipient in zip(donor.weights, recipient.weights):
        print(
            f"Checking: {w_donor.name}: {w_donor.shape} -> {w_recipient.name} : {w_recipient.shape}, shape changed: {w_donor.shape == w_recipient.shape}"
        )


def get_suffix(weight_name: str) -> str:
    """Get last 2 segments of a weight name."""
    return "/".join(weight_name.split("/")[-2::])


def get_weights(model: tf.keras.Model, weight_name: str) -> tf.Tensor:
    """Get weights by full name, endswith, or abbreviation."""

    # Full name matching
    matched = [w for w in model.weights if w.name == weight_name]

    # Ends with name matching
    if len(matched) == 0:
        matched = [w for w in model.weights if w.name.endswith(weight_name)]

    # Abbreviation matching
    if len(matched) == 0:
        matched = [
            w
            for w in model.weights
            if model.weights_abbreviations[weight_name] in w.name
        ]

    if len(matched) == 0:
        raise ValueError(f"Could not find weights for {weight_name}")
    if len(matched) > 1:
        raise ValueError(
            f"Found multiple weights: {[w.name for w in matched]} with similar name to {weight_name}, please be more specific."
        )
    return matched[0]


def copy_transplant(
    donor: tf.keras.Model, recipient: tf.keras.Model, weight_name: str
) -> None:
    """Transplant weights from donor to recipient in weights that has matching shape."""

    w_recipient = get_weights(recipient, weight_name)
    w_donor = get_weights(donor, weight_name)

    print(
        f"Transplanting: {w_donor.name}:{w_donor.shape} -> {w_recipient.name}: {w_recipient.shape}"
    )

    assert (
        w_donor.shape == w_recipient.shape
    ), f"Shapes don't match: {w_donor.shape} != {w_recipient.shape}"

    w_recipient.assign(w_donor)


@dataclass
class SurgeryPlan:
    """A surgery plan for removing shrink_rate amount of the units in `target_layer`.

    TODO: Support for enlargement surgery?
    """

    layer: str
    original_units: int
    shrink_rate: float

    def __post_init__(self):
        """Validate the surgery plan."""

        if not (0 < self.shrink_rate < 1):
            raise ValueError(
                f"Shrink rate must be between 0 and 1, got {self.shrink_rate}"
            )

        # Keeping `keep_n` units
        self.keep_n = int(self.original_units * (1 - self.shrink_rate))
        print(f"Keeping {self.keep_n} out of {self.original_units} original units")
        if self.keep_n == 0:
            raise ValueError(
                f"Shrink rate {self.shrink_rate} is too high, no units left."
            )

        # Generate indices to keep (shared across all weights)
        self.keep_idx = sorted(random.sample(range(self.original_units), self.keep_n))
        print(f"Keep indices are: {self.keep_idx}")

    def __repr__(self):
        return f"SurgeryPlan(target_layer={self.layer}, original_units={self.original_units}, damage={self.shrink_rate}, keep_idx={self.keep_idx}, keep_n={self.keep_n})"


def make_recipient(model, surgery_plan: SurgeryPlan, make_model_fn: Callable):
    """Make a recipient model according to surgery plan."""

    config = model.get_config()

    to_config_key = {
        "hidden": "h_units",
        "phonology": "p_units",
        "cleanup": "c_units",
    }

    config[to_config_key[surgery_plan.layer]] = surgery_plan.keep_n
    print(f"New config: {config}")
    return make_model_fn(**config)


class Surgeon:
    """A class for transplanting weights from one model to another according to surgery_plan.

    Args:
        surgery_plan: A surgery plan for the transplant (specifying where the damage happens).

    The `surgery_plan` specify where the shrinkage happens, it will move all related weights using `lesion_transplant()` method.
    For other weights that are not related to the surgery_plan, it will just copy them over using `copy_transplant()` method.

    Also see connectionist.models.PMSP.shrink_layer().

    Example:

    ```python

    from connectionist.surgery import *
    from connectionist.data import ToyOP

    data = ToyOP()
    donor_model = PMSP(tau=0.2, h_units=10, p_units=9, c_units=5)
    y = donor_model(data.x_train)  # for instantiating the weights

    # Create surgery plan and surgeon
    plan = SurgeryPlan(layer='hidden', original_units=10, shrink_rate=0.5)
    surgeon = Surgeon(surgery_plan=plan)

    # Create recipient model and transplant weights
    new_model = make_recipient(model=donor_model, surgery_plan=plan, make_model_fn=PMSP)
    new_model.build(input_shape=donor_model.pmsp._build_input_shape)
    surgeon.transplant(donor=donor_model, recipient=new_model)
    ```

    """

    def __init__(self, surgery_plan: SurgeryPlan) -> None:
        self.plan = surgery_plan

    @staticmethod
    def _validate_axis(
        w_donor: tf.Tensor,
        w_recipient: tf.Tensor,
        axis: List[int],
    ) -> None:

        if not 0 < len(axis) <= 2:
            raise ValueError(f"Axis must be of length 1 or 2, got {axis}")

        # Check non self-connecting weights shapes
        if len(axis) == 1 and len(w_donor.shape) > 1:
            match_ax = 1 - axis[0]
            if w_donor.shape[match_ax] != w_recipient.shape[match_ax]:
                raise ValueError(
                    f"In {w_donor.name}, shapes don't match on axis {match_ax}: {w_donor.shape=}, {w_recipient.shape=}"
                )

    def lesion_transplant(
        self,
        donor: tf.keras.Model,
        recipient: tf.keras.Model,
        weight_name: str,
        idx: List[int],
        axis: List[int],
    ) -> None:
        """Transplant weights from donor to recipient in the weights that requires shrinking."""

        w_recipient = get_weights(recipient, weight_name)
        w_donor = get_weights(donor, weight_name)
        self._validate_axis(w_donor, w_recipient, axis=axis)

        print(
            f"Transplanting: {w_donor.name}:{w_donor.shape} -> {w_recipient.name}: {w_recipient.shape}"
        )

        w = tf.identity(w_donor)  # Copy the weights
        for a in axis:
            w = tf.gather(
                w, indices=idx, axis=a
            )  # slice the weight in each connected axis
        w_recipient.assign(w)  # assign the weights to recipient

    def transplant(self, donor: tf.keras.Model, recipient: tf.keras.Model) -> None:
        """Transplant all the weights from donor to recipient model."""

        # Execute lesion transplant (Move weights and remove a subset of units)
        conn_locs = donor.connection_locs[self.plan.layer]
        weight_names = list(conn_locs.keys())
        axis = list(conn_locs.values())

        for name, ax in zip(weight_names, axis):
            self.lesion_transplant(
                donor=donor,
                recipient=recipient,
                weight_name=name,
                idx=self.plan.keep_idx,
                axis=ax,
            )

        # Execute simple transplant (Only move weights)
        all_weights_names = list(donor.weights_abbreviations.keys())
        remaining_weights = [
            name for name in all_weights_names if name not in weight_names
        ]
        for name in remaining_weights:
            copy_transplant(donor=donor, recipient=recipient, weight_name=name)



import numpy as np
import tensorflow as tf


class ToyOP:
    """Toy O-to-P dataset containing only 20 examples.

    - x_train: word representations, fixed across time.
    - y_train: letter sequence representations, changing across time.
    """

    letters = "crsaoutbn"
    words = [
        "cat",
        "cab",
        "can",
        "cot",
        "cob",
        "con",
        "cut",
        "cub",
        "rat",
        "ran",
        "rot",
        "rob",
        "rut",
        "rub",
        "run",
        "sat",
        "sot",
        "sob",
        "son",
        "sun",
    ]

    def __init__(self, max_ticks: int = 30):
        self.max_ticks = max_ticks
        self.x_train, self.y_train = self.get_data()

        if max_ticks % 3 != 0:
            raise ValueError("max_ticks must be divisible by 3")

    @property
    def n(self) -> int:
        return len(self.words)

    def get_data(self):
        word_repr = np.zeros((self.n, self.n))
        spelling_repr = np.zeros((self.n, len(self.words[0]), len(self.letters)))

        # Encode
        for i, word in enumerate(self.words):
            word_repr[i][i] = 1.0
            for j, letter in enumerate(word):
                spelling_repr[i][j][self.letters.index(letter)] = 1

        # Repeat over time axis
        x_train = np.stack([word_repr for _ in range(self.max_ticks)], axis=1)
        y_train = np.repeat(spelling_repr, self.max_ticks / 3, axis=1)

        return (
            tf.convert_to_tensor(x_train, dtype=tf.int8),
            tf.convert_to_tensor(y_train, dtype=tf.int8),
        )

    def __repr__(self) -> str:
        return f"Time invariant orthographic word representation: {self.x_train.shape=}.\nTime varying letter/phoneme sequence representation{self.y_train.shape=}"
